use termion::event::MouseEvent;

use crate::token::Token;

type Result<T> = std::result::Result<T, anyhow::Error>;

pub trait Widget {
    fn width(&self, max: usize) -> usize;
    fn line(&self, line: usize) -> Option<&Token>;
    fn on_event(
        &mut self,
        event: termion::event::Event,
    ) -> Result<()>;
}

pub enum WidgetWidth {
    Percent(u8),
    Absolute(usize),
}

impl WidgetWidth {
    pub fn absolute(&self, budget: usize) -> usize {
        match self {
            Self::Percent(p) => {
                // budget   = 100%
                // absolute = p%
                // absolute = (p * budget) / 100
                (*p as usize * budget) / 100
            }
            Self::Absolute(a) => *a,
        }
    }
}

pub struct ScrollableDisplay {
    lines: Vec<Token>,
    scrolled_offset: usize,
    width: WidgetWidth,
}

impl ScrollableDisplay {
    pub fn new(lines: Vec<Token>, width: WidgetWidth) -> Self {
        Self {
            lines,
            width,
            scrolled_offset: 0,
        }
    }
}

impl Widget for ScrollableDisplay {
    fn on_event(
        &mut self,
        event: termion::event::Event,
    ) -> Result<()> {
        let moving_value = match event {
            termion::event::Event::Key(_) => todo!(),
            termion::event::Event::Mouse(mus) => {
                if let MouseEvent::Press(mus, _, _) = mus {
                    match mus {
                        termion::event::MouseButton::WheelUp => {
                            (self.lines.len() - 1)
                                .min(self.scrolled_offset + 1)
                        }
                        termion::event::MouseButton::WheelDown => {
                            if self.scrolled_offset == 0 {
                                0
                            } else {
                                self.scrolled_offset - 1
                            }
                        }
                        _ => 0,
                    }
                } else {
                    0
                }
            }
            termion::event::Event::Unsupported(_) => 0,
        };
        if moving_value == 0 {
            return Ok(());
        }
        self.scrolled_offset = moving_value;
        Ok(())
    }

    #[inline(always)]
    fn width(&self, max: usize) -> usize {
        self.width.absolute(max)
    }

    fn line(&self, line: usize) -> Option<&Token> {
        self.lines.get(line + self.scrolled_offset)
    }
}
